/*
 * Kibitz - Searching the index
 *
 * (c) 2005-2014 Tomas Valla <tom@ucw.cz>
 *
 * This software may be freely distributed and used according to the terms
 * of the GNU Lesser General Public License. See COPYING for details.
 */

#undef LOCAL_DEBUG

#include <kibitz/lib.h>
#include <ucw/heap.h>
#include <ucw/io.h>
#include <ucw/unaligned.h>
#include <kibitz/internal.h>

static byte *index_ptr;
static size_t index_len;
static struct index_header *idx_header;
static byte *idx_phrases;
static u32 *idx_wordoffs;
static u32 *idx_fastfinds;
static byte *idx_trie;

static void scan_header(void)
{
	idx_header = (typeof(idx_header)) index_ptr;

	if (idx_header->magic == INDEX_TEMP_MAGIC)
		die("Temporary index only");
	if (idx_header->magic != INDEX_MAGIC)
		die("Invalid index (bad index magic)");

	idx_phrases = index_ptr + idx_header->phrase_start;
	idx_wordoffs = (u32 *) (index_ptr + idx_header->wordoff_start);
	idx_fastfinds = (u32 *) (index_ptr + idx_header->fastfind_start);
	idx_trie = index_ptr + idx_header->trie_start;

	uns v = idx_header->version;
	msg(L_INFO, "Loaded index version 0x%x, custom type 0x%x, custom version 0x%x",
		v >> 16, (v >> 8) & 0xff, v & 0xff);

}

static void touch_index(volatile byte *ptr, uns size)
{
	for (uns i=0; i<size; i += 4096)
		(void)ptr[i];
}

void kibitz_open_index(byte *idxname)
{
	if (index_ptr)
		munmap_file(index_ptr, index_len);
	index_ptr = mmap_file(idxname, &index_len, 0);
	touch_index(index_ptr, index_len);
	scan_header();
}

void kibitz_close_index(void)
{
	munmap_file(index_ptr, index_len);
}

union trie_result {
	struct {
		uns wref_index;
		uns wref_cnt;
		uns exact_cnt;
	} w;
	struct {
		uns ff_index;
		uns ff_cnt;
	} f;
};

/* returns: 0 if not found, <0 if result in ff_list, >0 if direct result */
static int trie_lookup(byte *word, uns wlen, union trie_result *result)
{
	byte *v = idx_trie;
	uns l = 0;
	byte *vstart = v;

	/* First jump through the vertices into the final one */
	while (l < wlen) {
		uns op = *v++;

		if (op == 0xff)		// vertex_end
			return 0;

		if ((op & 0xc0) == 0x80) {	// fastfind
			v += 4;
			continue;
		}

		if ((op & 0xf0) == 0x10) {	// skip_char
			byte *ve = v + (op & 15);
			while (l < wlen && v < ve && *v == word[l])
				v++, l++;
			if (l == wlen)
				break;
			if (v < ve)
				return 0;
			continue;
		}

		if ((op & 0xc0) == 0x40) {	// match
			if (op & 0x3f)		// small
				v += 4;
			else			// big
				v += 8;
			continue;
		}

		if ((op & 0xc0) == 0xc0) {	// exact match
			if (op & 0x1f)		// small
				;
			else			// big
				v += 4;
			continue;
		}

		if (op & 0x20) {		// branching
			if (op & 0x1f) {	// near
				uns c = op & 0x1f;

				/* FIXME: binsearch? */
				for (uns i=0; i<c; i++) {
					uns ch = *v++;
					uns roff = GET_U16(v);
					v += 2;
					if (ch == word[l]) {
						l++;
						vstart += roff;
						v = vstart;
						goto scan_end;
					}
				}
				return 0;
			}
			else {			// far
				uns c = *v++;

				/* FIXME: binsearch? */
				for (uns i=0; i<c; i++) {
					uns ch = *v++;
					uns off = GET_U32(v);
					v += 4;
					if (ch == word[l]) {
						l++;
						vstart = idx_trie + off;
						v = vstart;
						goto scan_end;
					}
				}
				return 0;
			}
			continue;
		}
		ASSERT(0);
scan_end:
	;
	}

	/* Now scan the final vertex */
	uns ff_index = 0, ff_cnt = 0;
	uns exact_cnt = 0, wref_cnt = 0, wref_index = 0;
	for (v = vstart;;) {
		uns op = *v++;

		if (op == 0xff)			// vertex_end
			break;

		if ((op & 0xc0) == 0x80) {		// fastfind
			ff_cnt = op & 0x3f;
			ff_index = GET_U32(v);
			v += 4;
		}

		if ((op & 0xf0) == 0x10) {		// skip_char
			byte *ve = v + (op & 15);
			while (l < wlen && v < ve && *v == word[l])
				v++, l++;
			if (l != wlen && v < ve)
				return 0;
			v = ve;
			continue;
		}

		if ((op & 0xc0) == 0x40) {		// match
			if (op & 0x3f) {		// small
				wref_cnt = op & 0x3f;
				wref_index = GET_U32(v);
				v += 4;
			}
			else {				// big
				wref_index = GET_U32(v);
				wref_cnt = GET_U32(v+4);
				v += 8;
			}
			continue;
		}

		if ((op & 0xc0) == 0xc0) {		// exact match
			if (op & 0x1f)
				exact_cnt = op & 0x1f;
			else {
				exact_cnt = GET_U32(v);
				v += 4;
			}
			continue;
		}
		break;
	}

	/* And forge the results */
	if (!(wref_cnt+ff_cnt))
		return 0;
	if (ff_cnt) {
		result->f.ff_cnt = ff_cnt;
		result->f.ff_index = ff_index;
		return -ff_cnt;
	}
	result->w.wref_cnt = wref_cnt;
	result->w.wref_index = wref_index;
	result->w.exact_cnt = exact_cnt;
	return wref_cnt;
}

static int build_ff_resbuf(union trie_result *tr, struct result resbuf[], uns rcnt)
{
	uns fi = tr->f.ff_index;
	uns fc = tr->f.ff_cnt;
	uns i;

	for (i=0; i<MIN(rcnt, fc); i++) {
		u32 ind = idx_fastfinds[2*(fi+i)];
		u32 q = idx_fastfinds[2*(fi+i) + 1];
		resbuf[i].phrase = (struct phrase *) (idx_phrases + ind);
		resbuf[i].quality = q;
	}
	return i;
}

#define addr_int_t intmax_t
struct hs {
	addr_int_t p;
};
#define HASH_PREFIX(x) ph_##x
#define HASH_NODE struct hs
#define HASH_KEY_ATOMIC p
#define HASH_ATOMIC_TYPE addr_int_t
#define HASH_WANT_CLEANUP
#define HASH_WANT_FIND
#define HASH_WANT_NEW
#define HASH_WANT_DELETE
#define HASH_AUTO_POOL 65536
#include <ucw/hashtable.h>

#define HEAP_LESS(x,y) ((x).quality < (y).quality)
static int build_wr_resbuf(union trie_result *tr, struct result resbuf[], uns rcnt)
{
	uns wi = tr->w.wref_index;
	uns wc = tr->w.wref_cnt;
	uns ec = tr->w.exact_cnt;
	struct result heap[rcnt+2];
	uns i, j;

	ph_init();

	for (i=0,j=0; i<rcnt+1 && j<wc; j++) {
		struct phrase *p = (struct phrase *) (idx_phrases + idx_wordoffs[wi + j]);
		int q = p->relevance;
		if (j < ec)
			q *= config.exact_match_bonus;

		if (ph_find((addr_int_t) p)) {
			for (uns k=1; k<=i; k++)
				if (heap[k].phrase == p) {
					heap[k].quality += q;
					break;
				}
		}
		else {
			ph_new((addr_int_t) p);
			heap[++i].phrase = p;
			heap[i].quality = q;
		}
	}
	HEAP_INIT(struct result, heap, i, HEAP_LESS, HEAP_SWAP);

	for (; j<wc; j++) {
		struct phrase *p = (struct phrase *) (idx_phrases + idx_wordoffs[wi + j]);
		int q = p->relevance;
		if (j < ec)
			q *= config.exact_match_bonus;

		if (ph_find((addr_int_t) p)) {
			for (uns k=1; k<=i; k++)
				if (heap[k].phrase == p) {
					heap[k].quality += q;
					HEAP_INCREASE(struct result, heap, i, HEAP_LESS, HEAP_SWAP, k, heap[k]);
					break;
				}
		}
		else {
			ph_delete((addr_int_t) heap[1].phrase);
			HEAP_DELETE_MIN(struct result, heap, i, HEAP_LESS, HEAP_SWAP);

			struct result n = {
				.phrase = p,
				.quality = q,
			};
			ph_new((addr_int_t) p);
			HEAP_INSERT(struct result, heap, i, HEAP_LESS, HEAP_SWAP, n);
		}
	}
	if (i == rcnt+1)
		HEAP_DELETE_MIN(struct result, heap, i, HEAP_LESS, HEAP_SWAP);

	uns r = i;
	while (i) {
		resbuf[i-1] = heap[1];
		HEAP_DELETE_MIN(struct result, heap, i, HEAP_LESS, HEAP_SWAP);
	}

	ph_cleanup();
	return r;
}

uns search_index(byte *word, uns wlen, struct result res[], uns max_res)
{
	union trie_result tr;

	int r = trie_lookup(word, wlen, &tr);
	DBG("trie_lookup for '%s' returned %d results", word, r);
	if (!r)
		return 0;
	if (r<0) {
#ifdef LOCAL_DEBUG
		for (uns i=0; i<tr.f.ff_cnt; i++) {
			uns woff = idx_fastfinds[2*(tr.f.ff_index+i)];
			uns q = idx_fastfinds[2*(tr.f.ff_index+i) + 1];
			struct phrase *p = (typeof(p)) (idx_phrases + woff);
			DBG("trie_result phrase='%s' q=%d", p->text, q);
		}
#endif
		r = build_ff_resbuf(&tr, res, max_res);
	}
	else {
#ifdef LOCAL_DEBUG
		for (uns i=0; i<tr.w.wref_cnt; i++) {
			struct phrase *p = (typeof(p)) (idx_phrases + idx_wordoffs[tr.w.wref_index+i]);
			DBG("trie_result wref=%d rel=%d phrase='%s' %s", tr.w.wref_index+i, p->relevance, p->text, (i < tr.w.exact_cnt ? "exact" : ""));
		}
#endif
		r = build_wr_resbuf(&tr, res, max_res);
	}

	return r;
}

